// VinciaShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the VinciaShower class 
  
#ifndef VINCIA_VinciaShower_H
#define VINCIA_VinciaShower_H

#include "TimeShower.h"
#include "VinciaAntenna.h"
#include "VinciaCommon.h"
#include "VinciaResolution.h"
#include "VinciaFunctions.h"
#include "VinciaMatching.h"
#include "VinciaTrialGenerators.h"

using namespace Pythia8;

// Declare the F77 3->2 clustering map for use below
extern"C" {
  void vc3to2_(int *MMAP, double PA[4], double PR[4], double PB[4],
	       double PAHAT[4], double PBHAT[4], int *IFAIL);
}

namespace Vincia {
  
  // Classes contained in this file (forward declarations)
  class Branchelemental;
  class VinciaShower;  
  
//**************************************************************************

// The VinciaShower class: main shower class for the timelike antenna shower   
// Inherits from TimeShower in Pythia 8 so can be used as alternative to 
// TimeShower.
  
  class VinciaShower : public Pythia8::TimeShower {
    
    // Allow VinciaPlugin to access private information
    friend class VinciaPlugin;
    
  public:
   
    // ----------------------------------------------------------------------
    
    // CONSTRUCTOR AND DESTRUCTOR 
    
    // Constructor.
    VinciaShower() : isInit(false), headerIsPrinted(false) {};
  
    // Destructor.
    virtual ~VinciaShower() {};
  
    // -----------------------------------------------------------------------
    
    // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING 
    
    // Initialize alphaStrong and related pTmin parameters.
    virtual void init( BeamParticle* beamAPtrIn = 0, 
		       BeamParticle* beamBPtrIn = 0);
    
    // Top-level routine to do a full time-like shower in resonance decay.
    virtual int shower( int iBeg, int iEnd, Event& event, double pTmax, 
			int nBranchMax = 0);
    
    // Prepare system for evolution.
    virtual void prepare( int iSys, Event& event, bool limitPTmaxIn,
                          double pTfirstTrialIn = 1e9);
    
    // Update dipole list after each ISR emission.  
    virtual void update( int iSys, Event& event);
    
    // Select next pT in downwards evolution.
    virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
                           bool isFirstTrial = false);
    
    // Branch event
    virtual bool branch( Event& event, bool isInterleaved=false); 
    
    // Update dipole list after a multiple interactions rescattering.
    virtual void rescatterUpdate( int iSys, Event& event);
    
    // Utility to print dipole list; for debug mainly.
    virtual void list( ostream& os = cout ) const;
    
    // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING              
    
    // -----------------------------------------------------------------------
    
    // PUBLIC NON-INHERITED MEMBERS
    
    // Initialize Pointers to Antenna sets
    void initAntPtr(AntennaSet* antUSR, AntennaSet* antDEF=0, 
		    AntennaSet* antMAX=0, AntennaSet* antMIN=0) {
      antennaSets.resize(0);
      antennaSets.push_back(antUSR);
      if (antDEF != 0 && antMAX != 0 && antMIN != 0) {
	antennaSets.push_back(antDEF);
	antennaSets.push_back(antMAX);
	antennaSets.push_back(antMIN);
      }
    }
    void initAntSctPtr(AntennaSet* antUSR, AntennaSet* antDEF=0,
                       AntennaSet* antMAX=0, AntennaSet* antMIN=0) {
      antennaSetsSct.resize(0);
      antennaSetsSct.push_back(antUSR);
      if (antDEF != 0 && antMAX != 0 && antMIN != 0) {
	antennaSetsSct.push_back(antDEF);
	antennaSetsSct.push_back(antMAX);
	antennaSetsSct.push_back(antMIN);
      }
    }
    
    //Initialize internal and diagnostic histograms
    void initVinciaHistos();
    
    // Print header information (version, settings, parameters, etc)
    void header();
    
    // Print list of winning branchings so far; for debug mainly.
    //virtual void history( ostream& os = cout ) {};
    
    // Print final statistics information
    void printInfo( ostream& os = cout );
    
    // Print internal and diagnostic histrograms
    void printHistos( ostream& os = cout ) {
      // Print histograms in matching class
      matchingPtr->printHistos(os);
      // Now print histograms in shower class
      map<string,Hist>::iterator iH;
      for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {	
	string Hname=iH->first;
        if ( vinciaHistos[Hname].getEntries() >= 1) {
	  os<<Hname<<vinciaHistos[Hname]<<endl;
	}
      }
    }
    
    // Print internal and diagnostic histrograms
    void writeHistos( string fileName="vincia", string lastName="dat" ) {
      // Print histograms in matching class
      matchingPtr->writeHistos( fileName, lastName );
      // Now print histograms in shower class
      map<string,Hist>::iterator iH;
      int hNumPhys = 0;
      int hNumTrial = 0;
      for (iH=vinciaHistos.begin();iH!=vinciaHistos.end();++iH) {	
	string Hname=iH->first;
        if ( vinciaHistos[Hname].getEntries() >= 1) {
          string file = fileName;
          if (Hname.find("Trial") != string::npos) {
            ++ hNumTrial;
            file += "-rTrial" + num2str(hNumTrial,int(log10(hNumTrial)+1));
          }
          else {
            ++ hNumPhys;
            file += "-rPhys" + num2str(hNumPhys,int(log10(hNumPhys)+1));
          }
          file += "." + lastName;
          cout<<" Writing "<<file<<endl;
          vinciaHistos[Hname].table(file);
	}
      }
    }
    
    // Allow the user to get a pointer to an internal histogram
    // Note: points to an entry of a map, so must be updated each
    // time it is used.
    Hist* getHistPtr(string histIn) {
      if (vinciaHistos.find(histIn) != vinciaHistos.end()) 
        return &vinciaHistos[histIn];
      else return 0;
    }
      
    // Kinematic maximum for gluon emission scale
    double yMaxEmit(int evType=0) {
      if (evType == 0) evType = evolutionType;
      if (evType == 1) return 1./pTnormalization;
      else if (evType == 2) return 1./mDnormalization;
      else return 1.;
    }

    // Evolution Windows: Phase space region boundaries
    int getRegion(double q) {
      if (q < mc/5) return 0;
      else if (q < mc) return 1;
      else if (q < mb) return 2;
      else if (q < mtb) return 3;
      else if (q < mt) return 4;
      else return int( log(q/mt)/log(5) + 5);
    }    

    // Evolution Windows: Phase space region boundaries
    double getQmin(int iRegion) {
      if (iRegion <= 0) return QminAll;
      else if (iRegion == 1) return mc/5.;
      else if (iRegion == 2) return mc;
      else if (iRegion == 3) return mb;
      else if (iRegion == 4) return mtb;
      else if (iRegion == 5) return mt;
      else return pow(5,(double)iRegion-5.0)*mt;
    }

    // Number of active flavors 
    int    getNf(int iRegion) {
      double qRegion = 1.1*getQmin(iRegion);
      return vinComPtr->getNf(qRegion);
    }

    // alpha_S:
    AlphaQCD* alphaSptr;        

    // Lambda value
    double getLambda(int nFin) {
      if (nFin <= 3) return Lambda3flav;
      else if (nFin <= 4) return Lambda4flav;
      else if (nFin <= 5) return Lambda5flav;
      else return Lambda6flav;
    }

    // Functions to set and get headRoomFactors
    void setHeadRoomEmit(double val, int nGin=-1, int nQin=-1) {
      // Initialize all headRoomFactors
      if (nGin == -1 && nQin == -1) {
        for (nGin = 0; nGin<9; ++nGin) 
          for (nQin = 0; nQin<9; ++nQin) {
            headRoomEmit[nGin][nQin] = val;
            headRoomEmitOptimize[nGin][nQin] = true;
          }
      }
      // Set specific headRoomFactor
      else if (nGin >= 0 && nGin < 9 && nQin >= 0 && nQin < 9) {
        // Switch off further optimization if increasing headroom factor
        // (Increase to a maximum of factor 2)
        if ( val > headRoomEmit[nGin][nQin] ) {
          headRoomEmitOptimize[nGin][nQin] = false;
          headRoomEmit[nGin][nQin] = min(val,2.0);
          if (verbose >= 5) {
            printOut("setHeadRoomEmit","Increasing headroom at nG = "
                     +num2str(nGin,2)+" nQ = "+num2str(nQin,2)+"; current "
                     "headRoom = "+num2str(headRoomEmit[nGin][nQin]));
          }
        } 
        // Allow reduction of headroom if we're still optimizing
        else if ( headRoomEmitOptimize[nGin][nQin] ) {
          headRoomEmit[nGin][nQin] = val;
        }
      }
    }
    void setHeadRoomSplit(double val, int nGin=-1, int nQin=-1) {
      // Initialize all headRoomFactors
      if (nGin == -1 && nQin == -1) {
        for (nGin=0; nGin<9; ++nGin) 
          for (nQin=0; nQin<9; ++nQin) {
            headRoomSplit[nGin][nQin] = val;
            headRoomSplitOptimize[nGin][nQin] = true;
          }
      }

      // Set specific headRoomFactor
      else if (nGin >= 0 && nGin < 9 && nQin >= 0 && nQin < 9) {
        // Switch off further optimization if increasing headroom factor
        // (Increase to a maximum of factor 4)
        if ( val > headRoomSplit[nGin][nQin] ) {
          headRoomSplitOptimize[nGin][nQin] = false;
          headRoomSplit[nGin][nQin] = min(val,4.0);
          if (verbose >= 5) {
            printOut("setHeadRoomSplit","Increasing headroom at nG = "
                     +num2str(nGin,2)+" nQ = "+num2str(nQin,2)+"; current "
                     "headRoom = "+num2str(headRoomSplit[nGin][nQin]));
          }
        } 
        // Allow reduction of headroom if we're still optimizing
        else if ( headRoomSplitOptimize[nGin][nQin] ) {
          headRoomSplit[nGin][nQin] = val;
        }
      }
    }
    double getHeadRoomFactorEmit(int nGin, int nQin) {
      if (nGin >= 0 && nGin < 9 && nQin >= 0 && nQin < 9) 
        return headRoomEmit[nGin][nQin];
      else 
        return 1.0;
    }
    double getHeadRoomFactorSplit(int nGin, int nQin) {
      if (nGin >= 0 && nGin < 9 && nQin >= 0 && nQin < 9) 
        return headRoomSplit[nGin][nQin];
      else 
        return 1.0;
    }

    // Add trial functions to a Branchelemental
    void resetTrialGenerators(Branchelemental* trial);

    // Get weight of current event; 
    // Normally 1 for set 0 (user settings) 
    // Use iVar > 0 to access uncertainty weights
    double weight(int iVar=0) const {
      if (iVar >= 0 && iVar < nVariations()) 
	return min(weightLimit*weightSave[0],weightSave[iVar]) ;
      else return 0.0 ;
    }   

    // Return number of weights available
    int nVariations() const { return weightSave.size() ;}  

    // Warning flags: set true if w < 0 and/or |w| > 1 occurred, respectively 
    bool reweightingOccurred;
    
    // Wrapper function to return a specific antenna
    // inside a specific antenna set
    // 0: User, 1: Def, 2: Max, 3: Min    
    Antenna* getAnt(int iAnt, bool isSectorIn=false, int iSet=0) {
      if (!isSectorIn) 
        return antennaSets[iSet]->getAnt(iAnt);
      else 
        return antennaSetsSct[iSet]->getAnt(iAnt);
    }
    
    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

  private:
    
    // -----------------------------------------------------------------------
    
    // PRIVATE MEMBERS
    
    // Main trial accept function
    bool acceptTrial(Event& event);

    //--------------------------------------------------------------------
    // AUXILIARIES to acceptBranching():

    // Veto function for sector ordering.
    bool sectorAccept(Event& event, Branchelemental* trialPtr);
    
    // Main matching routine, returns bool = accept (true) / reject (false)
    bool matchingAccept(Event& event, Branchelemental* trialPtr);

    //--------------------------------------------------------------------
    
    // Check if particular weight exists
    bool existsWeight(int iVar=0) {
      if (iVar >= 0 && iVar < nVariations()) return true;
      else return false;
    }

    // Set all event weights (e.g., for initialization)
    void setWeightAll( double weightIn ) {
      for (int iVar=0; iVar < nVariations(); iVar++) {
	setWeight(weightIn,iVar);      
      }
    }
    
    // Set all weight sums (e.g., for initialization)
    void resetWeightSums(double weightSumIn = 0.0) {
      for (int iVar=0; iVar < nVariations(); iVar++) {
	weightSumSave[iVar]=weightSumIn;
      }
    }
    
    // Set event weight, check for negative and non-unity weights.
    bool setWeight( double weightIn, int iVarIn=0);
    
    // Scale event weight
    bool scaleWeight( double scaleFacIn, int iVarIn=0) {
      if (existsWeight(iVarIn)) {
	weightSave[iVarIn] *= scaleFacIn;	
	return true;
      } else return false;
    }

    // Initialized?
    bool isInit;

    // AlphaS parameters 
    int alphaSorder, alphaSmode; 
    bool alphaScmw, alphaScmwDef;
    double alphaSvalue, alphaSkMu, alphaSmax, alphaSmuMin, alphaSkMuDef;
    double mt, mtb, ms, mb, mc;
    double Lambda3flav, Lambda4flav, Lambda5flav, Lambda6flav;
    double Lambda3flav2, Lambda4flav2, Lambda5flav2, Lambda6flav2;    

    // Shower parameters
    bool useScale;
    int evolutionType, cutoffType, cutoffMode, orderingMode, nGluonToQuark;
    double cutoffScale, QminAll;
    double pTnormalization, mDnormalization;
    bool normalizeInputPT;

    // Trial generators
    TrialSoft trialSoft;
    TrialCollI trialCollI;
    TrialCollK trialCollK;
    TrialSplitI trialSplitI;
    TrialSplitK trialSplitK;

    // Uncertainty Bands
    bool uncertaintyBands;
    double uncertaintyBandsKmu;

    // Pointer to Evolution Variable
    Resolution* resolutionPtr;

    // Pointer to VINCIA common function classes
    VinciaCommon* vinComPtr;
    Functions* functionsPtr;

    // Sector 
    int sectorType;
    bool sectorShower, useSectorTerms;

    // Polarization global flag
    bool helicityShower;
    // Polarization and mass local flags
    bool isMassless, usePolarizationNow;
    
    // Hyperjet on/off
    bool hyperjet;
    
    // Pointer to VINCIA clustering map
    VinClu* vinCluPtr;
    
    // Desired matching multiplicities, level, regulator, regulator scales
    Matching* matchingPtr;
    MGInterface* mgInterfacePtr;
    bool matching, isResonance;
    int  matchingLO, matchingNLO;
    int  matchingRegOrder, matchingRegShape, matchingRegType;
    double matchingScale, matchingScaleRatio;
    bool  matchingScaleIsAbs, matchingFullColor;
    vector<double> PME;

    // New headRoomFactors
    double headRoomEmit [9][9]; 
    double headRoomSplit [9][9];
    bool headRoomEmitOptimize [9][9]; 
    bool headRoomSplitOptimize [9][9];
    
    // Bool to tell whether MadGraph linked or not.
    bool isMGOn;    // True if real MG interface linked, false if dummy
    int nG, nQQ;    // Count numbers of quarks and gluons
    int MEClass;         // Code telling which originating resonance / process
    double MESav, MENew; // Storage of ME at previous matched order
    int MESavN;          // Number of partons at previous matched order
    // int nEmitPhoton, nSplitPhoton, nEmitZ, nSplitZ, nEmitW, nSplitW;
    
    // Pointers to the two incoming beams.
    BeamParticle* beamAPtr;
    BeamParticle* beamBPtr;
    
    // Vector of antenna sets (0 is physical, >= 1 are variations)
    vector<AntennaSet*> antennaSets;   
    vector<AntennaSet*> antennaSetsSct; 
    vector<double> Paccept;

    // vector of dipoles (with trial branchings) 
    vector<Branchelemental> branchelementals;

    // current winner
    Branchelemental* winnerPtr;
    int indxWin;

    // Store history of winners (e.g., to check alternative orderings)
    vector<Branchelemental> winners;

    // current global restart scale for unordered antennae
    double Qrestart;

    // current global ordering scale for unordered antennae
    double Q2UnOrdOld, mD2UnOrdOld;

    // Internal histograms
    map<string,Hist> vinciaHistos;
    
    // Total mass of showering system (used for Type 3 ordering)
    double mSystem;
    
    // Event weights
    vector<double> weightSave, weightSumSave, weightSum2Save;
    vector<double> weightOld, weightMaxSave;    
    double nUnityWeight, nNonunityWeight, nNegativeWeight;
    double nTotWeights, nWeights, weightLimit;
    double reweightMaxPSaveP, reweightMaxPSaveQ;  
    double reweightMaxQSaveP, reweightMaxQSaveQ;
    bool allowReweighting;
    
    // Statistics
    int nTrialsSum, nTrialsCutM, nTrialsCutQ, nTrialsVeto, nTrialsHull, nTrialsHullMassive;

    // Verbose setting and version number
    int verbose;
    double version;
    bool headerIsPrinted;    

  };
  
  //**************************************************************************
  
  // Auxiliary classes used by VinciaShower
  
  //**************************************************************************
  

  //**************************************************************************
  
  // The Branchelemental class: contains a generic 2->3 branching, with 
  // pointers to the relevant places in the event record and a few useful 
  // functions. 
  // The member function list() prints a brief record of the contents.
  // The member function setTrial(bool) is used to (re)set whether this
  // elemental already contains a saved trial branching, or whether a new 
  // one should be generated.
  
  class Branchelemental {
    
  public:
    
    // Constructors
    // Create empty branchelemental
  Branchelemental() : system(0) {}
    // Create branchelemental with iOld1In,iOld2In as parents
    // and initialize the PDT pointers for the new ones
  Branchelemental(Event& event, int iOld1In, int iOld2In)
    : system(0) {reset(event, iOld1In, iOld2In);};
    
    // Antenna Mass (negative if spacelike virtual)
    double mCalc() {
      sSav = m2(old1,old2);
      mSav = (sSav >= 0) ? sqrt(sSav) : -sqrt(-sSav);
      return mSav;
    }
    double m() const { return mSav; }
    double s() const { return sSav; }
    // CM energy squared of daughters (for cross checks)
    double sNew() const {return m2(new1,new2,new3);}
    // Dot products of daughters
    double s12() const {return m2(new1,new2) - new1.m2() - new2.m2();} 
    double s23() const {return m2(new2,new3) - new2.m2() - new3.m2();} 
    double s13() const {return m2(new1,new3) - new1.m2() - new3.m2();} 
    double sNew1() const {return new1.m2();} 
    double sNew2() const {return new2.m2();}
    double sNew3() const {return new3.m2();}

    // Reset and set trial true/false
    void reset(Event& event, int iOld1In, int iOld2In) {
      // Make local copy of input particles
      old1        = event[iOld1In];
      old2        = event[iOld2In];
      // Compute and store antenna invariant mass
      mCalc();      
      // Indices of parents and neighbors
      iOld1       = iOld1In;
      iOld2       = iOld2In;
      iL          = 0; 
      iR          = 0;
      iLL         = 0;
      iRR         = 0;
      // Trial Generators
      clearTrialGenerators();
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
      // Default antenna properties
      Qold        = 0.0;
      unOrdering  = false;
      new1=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new2=Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      new3=Particle(0,51,iOld2,iOld2,0,0,0,0,0.);      
      // Set pointers
      new1.setEvtPtr(&event);
      new2.setEvtPtr(&event);
      new3.setEvtPtr(&event);
    }

    // Function to reset all trial generators for this branchelemental
    void clearTrialGenerators() {
      trialGenPtrSav.resize(0);
      iAntPhysSav.resize(0);
      isSwappedSav.resize(0);
      hasSavedTrial.resize(0);
      scaleSav.resize(0);
      zMinSav.resize(0);
      zMaxSav.resize(0);
      colFacSav.resize(0);
      alphaSav.resize(0);
      headRoomSav.resize(0);
      nVeto       = 0;
      nHull       = 0;
      nHadr       = 0;
    }

    // Add a trial generator to this branchelemental. Return name of generator.
    int addTrialGenerator(int iAntPhysIn, bool swapIn,  
                          TrialGenerator* trialGenPtrIn) {
      trialGenPtrSav.push_back(trialGenPtrIn);
      iAntPhysSav.push_back(iAntPhysIn);
      isSwappedSav.push_back(swapIn);
      hasSavedTrial.push_back(false);      
      scaleSav.push_back(-1.0);
      zMinSav.push_back(0.0);
      zMaxSav.push_back(0.0);
      colFacSav.push_back(0.0);
      alphaSav.push_back(0.0);
      headRoomSav.push_back(1.0);
      return int(trialGenPtrSav.size()-1);
    }
    
    // Function to return number of trial generators for this antenna
    int nTrialGenerators() const { return trialGenPtrSav.size(); }

    // Save a generated trial branching
    void saveTrial(int iTrial,double qTrial,double zMin,double zMax,
                   double colFac,double alphaEff, double headRoom = 1.0) {
      hasSavedTrial[iTrial] = true;
      scaleSav[iTrial]      = qTrial;
      zMinSav[iTrial]       = zMin;
      zMaxSav[iTrial]       = zMax;
      colFacSav[iTrial]     = colFac;
      alphaSav[iTrial]      = alphaEff;
      headRoomSav[iTrial]   = headRoom;
    }

    // Generate invariants for saved branching
    bool genTrialInvariants(double& y12, double& y23, int iTrial = -1) {
      // Automatically determine which trial function to use if -1 input
      int iGen = iTrial;
      if (iGen == -1) iGen = getTrialIndex();
      if (iGen <= -1) return false; 
      double z = trialGenPtrSav[iGen]->genZ(zMinSav[iGen],zMaxSav[iGen]);
      // Check physical phase space (note: this only checks massless hull)
      // (Use absolute z value since negative z values are used to indicate 
      //  swapped invariants for mD ordering)
      double yE = pow2(scaleSav[iGen]/mSav);
      if (abs(z) < trialGenPtrSav[iGen]->getZmin(yE) || 
          abs(z) > trialGenPtrSav[iGen]->getZmax(yE)) return false;
      // Convert to y12 , y23
      y12 = trialGenPtrSav[iGen]->getYij(yE,z);
      y23 = trialGenPtrSav[iGen]->getYjk(yE,z);
      return true;
    }

    // Get trial function index of winner
    // (return -1 if not all trial generators have saved trials)
    int getTrialIndex() const {
      double qMax = 0.0;
      int iMax    = -1;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (!hasSavedTrial[i]) return -1;
        double qSav = scaleSav[i];
        if (qSav > qMax) {
          qMax = qSav;
          iMax = i;
        }
      }
      return iMax;
    }
        
    // Check if a saved trial exists for a particular trialGenerator
    bool hasTrial(int iTrial) const {
      if (iTrial < int(hasSavedTrial.size())) return hasSavedTrial[iTrial];
      else return false;
    }

    // Get whether physical antenna is swapped 
    bool getIsSwapped(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return isSwappedSav[iTrial];
    }
    
    // Get physical antenna function index of winner
    int getPhysIndex(int iTrial=-1) const {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return iAntPhysSav[iTrial];
    }
    
    // Get scale for a specific saved trial
    double getTrialScale(int iTrial) const {
      if (iTrial < int(scaleSav.size())) return scaleSav[iTrial];
      else return -1.0;
    }

    // Get scale of winner
    double getTrialScale() const {
      double qMax = 0.0;
      for (int i = 0; i < int(scaleSav.size()); ++i) {
        if (hasSavedTrial[i]) qMax = max(qMax,scaleSav[i]);
        else printOut("getTrialScale","Error! "
                      "not all trials have saved scales");
      }
      return qMax;
    }

    // Get color factor 
    double getColFac(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return colFacSav[iTrial];
    }
    
    // Get whether antenna is swapped 
    double getHeadRoomFactor(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return headRoomSav[iTrial];
    }
    
    // Get whether antenna is swapped 
    double getAlphaTrial(int iTrial=-1) {
      if (iTrial <= -1) iTrial = getTrialIndex();
      return alphaSav[iTrial];
    }
    
    // Flag to set if a saved trial should be ignored and a new one generated
    // Default value -1 : force all trials to renew
    void renewTrial(int iTrial = -1) {
      if (iTrial >= 0) hasSavedTrial[iTrial] = false;
      else {
        for (iTrial = 0; iTrial < int(hasSavedTrial.size()); ++iTrial)
          hasSavedTrial[iTrial] = false;
      }
    }

    // List function
    void list(ostream& os=cout, bool header=false, bool footer=false) const; 
    
    // Data storage members 
    int iOld1, iOld2;    
    Particle old1, old2;
    Particle new1, new2, new3;  
    // Nearest-Neighbour and Next-to-Nearest Neighbour indices
    int iL,iR;
    int iLL,iRR;
    // 
    int system,nVeto,nHull,nHadr;
    double Qold;
    // Ultimate ancestor (e.g., if part of decay system)
    int iAncestor;
    // Has trial branching, and whether strong ordering should be imposed
    bool unOrdering;  

    // Trial Generators and properties of saved trials
    vector<TrialGenerator*> trialGenPtrSav;
    vector<double> zMinSav, zMaxSav, colFacSav, alphaSav;
    vector<double> scaleSav, headRoomSav;
    vector<bool> hasSavedTrial, isSwappedSav;
    vector<int> iAntPhysSav;

   private:

    // Saved antenna invariant mass value
    double mSav, sSav;

  };
  
}

#endif
