// VRfile.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2013 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VRfile class 

// Include the VRfile header
#include "VRfile.h"

namespace Vincia {

  // Read data points from file
  bool VRfile::readFile(string fileNameIn, int verboseIn) {
  
    // Initialize
    init(verboseIn);
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRfile): begin read "<<fileNameIn<<endl;
    
    // Open input stream. Return false if fails
    ifstream inStream;
    inStream.open(fileNameIn.c_str());
    if (! inStream) return false;
    
    // Read lines, return false if no useful lines
    int nLines = 0;
    while (1) {
      if ( readLine(inStream)) ++nLines; 
      if ( !inStream.good() ) break;
    }
    if (nLines <= 0) {
      if (verbose >= 1) 
        cout<<"(VRfile): warning! no readable lines in "<<fileNameIn<<endl;
      return false;
    }
    // Verbose output
    if (verbose >= 2) {
      cout<<"(VRfile): "<<nLines<<" readable data points in "
          <<fileNameIn<<endl;
    }
    
    // Read SRC, REF, and ECM if present
    inStream.clear();
    char line[256];
    while (1) {
      inStream.getline(line,256);
      if (inStream.good()) {
        string sline(line);                    
        int ipad = 1;
        if (sline.substr(sline.find("=")+1,1) == " ") ipad = 2;
        if (sline.find("SRC") != string::npos) {
          fileSRC = sline.substr(sline.find("=")+ipad,sline.size());
        } else if (sline.find("REF") != string::npos) {
          fileREF = sline.substr(sline.find("=")+ipad,sline.size());
        } else if (sline.find("SQRT(S)") != string::npos ||
                   sline.find("ECM") != string::npos ) {
          fileECM = sline.substr(sline.find(":")+ipad,sline.size());
        }
      } else {
        break;
      }
    }
    
    // Finalize read-in (compute sum of weights, etc)
    bool ok = finalize();
    
    // Everything ok?
    return ok;
  }

  // Read line from data file. 
  bool VRfile::readLine(ifstream& inStream) {

    // Sanity check
    if (format.size() == 0) return false;
    
    // Read values into buffer (default: assume unit-wide bins)
    double val, xL(0.5), xC, xR(0.5), y(0.);
    double sigStatHi(0.), sigStatLo(0.), sigSysHi(0.), sigSysLo(0.);
    double sigTotHi(0.), sigTotLo(0.), sigSys2Hi(0.), sigSys2Lo(0.);
    bool hasX       = false;
    bool hasDX      = true;
    bool hasStatErr = false;
    bool hasSysErr  = false;
    bool hasSysErr2 = false;
    bool hasTotErr  = false;      
    for (unsigned int iFmt=0; iFmt<format.size(); ++iFmt) { 
      inStream >> val;
      if (!inStream.good()) {
        if (format[iFmt] <= 0) continue;
        else return false;
      }
      if (format[iFmt] <= 0) continue;
      if (format[iFmt] >= 10 && format[iFmt] <= 19) {
        if (format[iFmt] == 10 || format[iFmt] == 12) hasDX = false;
        if (format[iFmt] <= 11) xL = val;
        else if (format[iFmt] == 12) xR = 1.0 - val;
      } 
      else if (format[iFmt] >= 20 && format[iFmt] <= 29) {
        hasX = true;
        xC = val;
        if (format[iFmt] == 22) xC = 1.0 - val;
      } 
      else if (format[iFmt] >= 30 && format[iFmt] <= 39) {
        if (format[iFmt] == 30) hasDX = false;
        if (format[iFmt] <= 31) xR = val;
        else if (format[iFmt] == 32) xL = 1.0 - val;
      }
      else if (format[iFmt] >= 40 && format[iFmt] <= 49) {
        y  = val;
      }
      else if (format[iFmt] >= 50 && format[iFmt] <= 59) {
        hasStatErr = true;
        if (format[iFmt] == 50 || format[iFmt] == 51) sigStatLo = val;
        if (format[iFmt] == 50 || format[iFmt] == 52) sigStatHi = val;
      }
      else if (format[iFmt] >= 60 && format[iFmt] <= 69) {
        hasSysErr  = true;
        if (format[iFmt] == 60 || format[iFmt] == 61) sigSysLo = val;
        if (format[iFmt] == 60 || format[iFmt] == 62) sigSysHi = val;
      }
      else if (format[iFmt] >= 70 && format[iFmt] <= 79) {
        hasSysErr2 = true;
        if (format[iFmt] == 70 || format[iFmt] == 71) sigSys2Lo = val;
        if (format[iFmt] == 70 || format[iFmt] == 72) sigSys2Hi = val;
      }
      else if (format[iFmt] >= 80 && format[iFmt] <= 89) {
        hasTotErr  = true;
        if (format[iFmt] == 80 || format[iFmt] == 81) sigTotLo  = val;
        if (format[iFmt] == 80 || format[iFmt] == 82) sigTotHi  = val;
      }
    }
    
    // If dxL and dxR given, construct xL and xR from xC
    if (hasDX) {
      if (!hasX) cout<<"(VRfile): cannot construct bin edges"
                   " without central X value"<<endl;
      xL = xC - abs(xL);
      xR = xC + abs(xR);
    }
    // If only xL and xR given, construct xC
    if (!hasX) {
      xC = (xL + xR)/2.;
    }
    // If only 2nd sys error given, move to 1st
    if (hasSysErr2 && !hasSysErr) {
      hasSysErr2 = false;
      hasSysErr  = true;
      sigSysLo   = sigSys2Lo;
      sigSysHi   = sigSys2Hi;
      sigSys2Lo  = 0.;
      sigSys2Hi  = 0.;
    }
    // Construct total error, summing stat and sys errors in quadrature but
    // individual sys errors linearly if 2 separate ones given
    if (!hasTotErr) {
      sigTotLo = sqrt(pow(sigStatLo,2)+pow(abs(sigSysLo)+abs(sigSys2Lo),2));
      sigTotHi = sqrt(pow(sigStatHi,2)+pow(abs(sigSysHi)+abs(sigSys2Hi),2));
    }
    // Construct missing stat error if necessary
    if (hasTotErr && !hasStatErr) {
      double sigSysSumLo = abs(sigSysLo) + abs(sigSys2Lo);
      double sigSysSumHi = abs(sigSysLo) + abs(sigSys2Lo);
      sigStatLo = sqrt(max(0.0,pow(sigTotLo,2)-pow(sigSysSumLo,2)));
      sigStatHi = sqrt(max(0.0,pow(sigTotHi,2)-pow(sigSysSumHi,2)));
    }

    // Optionally suppress zero values with zero uncertainty
    if ( y == 0.0 && sigTotLo == 0.0 && sigTotHi == 0.0 ) return true;
    
    // Add data point
    // assume x values are ordered ascending or descending, find out which
    if (xCsav.empty() || xC < *(xCsav.begin())) {
      // insert at the beginning
      xLsav.insert(xLsav.begin(), xL);
      xCsav.insert(xCsav.begin(), xC);
      xRsav.insert(xRsav.begin(), xR);
      ySav.insert(ySav.begin(), y);
      uTotLoSav.insert(uTotLoSav.begin(), sigTotLo);
      uTotHiSav.insert(uTotHiSav.begin(), sigTotHi);
      uStatLoSav.insert(uStatLoSav.begin(), abs(sigStatLo));
      uStatHiSav.insert(uStatHiSav.begin(), abs(sigStatHi));
      uSysLoSav.insert(uSysLoSav.begin(), abs(sigSysLo));
      uSysHiSav.insert(uSysHiSav.begin(), abs(sigSysHi));
      uSys2LoSav.insert(uSys2LoSav.begin(), abs(sigSys2Lo));
      uSys2HiSav.insert(uSys2HiSav.begin(), abs(sigSys2Hi));
    } else {
      // insert at the end
      xLsav.push_back(xL);
      xCsav.push_back(xC);
      xRsav.push_back(xR);
      ySav.push_back(y);
      uTotLoSav.push_back(sigTotLo);
      uTotHiSav.push_back(sigTotHi);
      uStatLoSav.push_back(abs(sigStatLo));
      uStatHiSav.push_back(abs(sigStatHi));
      uSysLoSav.push_back(abs(sigSysLo));
      uSysHiSav.push_back(abs(sigSysHi));
      uSys2LoSav.push_back(abs(sigSys2Lo));
      uSys2HiSav.push_back(abs(sigSys2Hi));
    }
    
    
    // Store number of systematic uncertainties
    nSysSav = 0;
    if (hasSysErr) ++nSysSav;
    if (hasSysErr2) ++nSysSav;
    
    // Line read ok
    return true;
  }
  
  // Finalize read-in: fill in empty arrays and compute normalization.
  // Assumes x values are ascending
  bool VRfile::finalize() {
    
    // Zero weight sum
    sumWeights = 0.;
    
    // Loop over data points        
    for (int i=0; i<int(ySav.size()); ++i) {
      // Check for overlapping bins and restrict if necessary
      if (i+1 < int(ySav.size())) {
        if (xRsav[i] > xLsav[i+1]) {
          xRsav[i]   = 0.5 * (xCsav[i] + xCsav[i+1]);
          xLsav[i+1] = xRsav[i];
          if (i==0) xLsav[i]   = xCsav[i] - (xRsav[i] - xCsav[i]);
          if (i+1==int(ySav.size())-1) 
            xRsav[i+1] = xCsav[i+1] + (xCsav[i+1] - xLsav[i+1]);
        }
      }
      double dx = (xRsav[i]-xLsav[i]);
      sumWeights += ySav[i] * dx;
    }
    
    // Everything ok.
    return true;
  }
  
} // end namespace VINCIA
